################################################################################
#                                                                              #
#                           DARTFlo viscous module file                        #
#                 Transition solver : Transition related components.           #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
from dart.viscous.Physics.DClosures import Closures

import numpy as np

class Transition:
    def __init__(self, xtrF):
        self.upperForced = xtrF[0]
        self.lowerForced = xtrF[1]

    def initTransition(self, var):

        var.flowRegime=np.zeros(var.nN)
        var.flowRegime[var.bNodes[2]:]=1
        
        for n in range(len(var.xtrF)):
            if var.xtrF[n] is not None:
                if var.xtrF[n]==1: # Forced laminar flow
                    var.transMarkers[n]=var.bNodes[n+1]
                    var.xtr[n]=1
                    var.flowRegime[var.bNodes[n]:var.bNodes[n+1]]=0
                elif var.xtrF[n]==0: # Forced turbulent flow
                    var.transMarkers[n]=var.bNodes[n]
                    var.xtr[n]=0
                    var.flowRegime[var.bNodes[n]:var.bNodes[n+1]]=1
                elif 0<var.xtrF[n]<1:
                    idx0=(abs(var.xGlobal[n*var.bNodes[1]:var.bNodes[1+n]]-0)).argmin()
                    var.transMarkers[n]=(abs(var.xGlobal[n*var.bNodes[1]+idx0:var.bNodes[1+n]]
                    -var.xtrF[n])).argmin()+n*var.bNodes[1]+idx0
                    var.flowRegime[var.transMarkers[n]:var.bNodes[n+1]]=1
                    var.xtr[n]=var.xtrF[n]
                else:
                    raise ValueError('Non-physical forced transition location')    


    def solveTransition(self,dataCont,transMarker):
      
      dataCont.flowRegime[dataCont.bNodes[2]:] = 1		# Wake 

      # Upper side
      if dataCont.bNodes[0]<transMarker<dataCont.bNodes[1]:
          
        # Set regime to turbulent on remaining upper side points
        dataCont.transMarkers[0] = transMarker
        dataCont.flowRegime[transMarker : dataCont.bNodes[1]] = 1

        # Compute upper side transition location
        dataCont.xtr[0] = (dataCont.Ncrit-dataCont.u[(transMarker-1)*dataCont.nVar+2])*((dataCont.xGlobal[transMarker]-dataCont.xGlobal[transMarker-1])/(dataCont.u[transMarker*dataCont.nVar+2]-dataCont.u[(transMarker-1)*dataCont.nVar+2]))+dataCont.xGlobal[transMarker-1]
        avTurb = (dataCont.xGlobal[transMarker]-dataCont.xtr[0])/(dataCont.xGlobal[dataCont.transMarkers[0]]-dataCont.xGlobal[dataCont.transMarkers[0]-1]) # percentage of the subsection corresponding to a turbulent flow
    
      # Fully turbulent flow on lower side
      elif transMarker == dataCont.bNodes[1]:
        dataCont.transMarkers[1] = dataCont.bNodes[1]
        dataCont.xtr[1] = 0
        dataCont.flowRegime[dataCont.bNodes[1] : dataCont.bNodes[2]] = 1

      # Lower side
      elif dataCont.bNodes[1] <= transMarker < dataCont.bNodes[2]:

        # Set regime to turbulent on remaining lower side points
        dataCont.transMarkers[1] = transMarker
        dataCont.flowRegime[transMarker : dataCont.bNodes[2]] = 1

        # Compute lower side transition location
        dataCont.xtr[1] = (dataCont.Ncrit - dataCont.u[(transMarker-1)*dataCont.nVar+2])*((dataCont.xGlobal[transMarker]-dataCont.xGlobal[transMarker-1])/(dataCont.u[transMarker*dataCont.nVar+2]-dataCont.u[(transMarker-1)*dataCont.nVar+2])) + dataCont.xGlobal[transMarker-1]
        avTurb = (dataCont.xGlobal[transMarker] - dataCont.xtr[1])/(dataCont.xGlobal[dataCont.transMarkers[1]]-dataCont.xGlobal[dataCont.transMarkers[1]-1]) # percentage of the subsection corresponding to a turbulent flow
    
      # Boundary condition
      avLam = 1- avTurb # percentage of the subsection corresponding to a laminar flow
      dataCont.blPar[(transMarker-1)*dataCont.nBlPar + 7] = Closures.turbulentClosure(dataCont.u[(transMarker-1)*dataCont.nVar+0], dataCont.u[(transMarker-1)*dataCont.nVar+1],0.03, dataCont.blPar[(transMarker-1)*dataCont.nBlPar+0], dataCont.extPar[(transMarker-1)*dataCont.nExtPar+0],'airfoil')[6]
      dataCont.u[(transMarker-1)*dataCont.nVar + 4] = 0.7*avLam*dataCont.blPar[(transMarker-1)*dataCont.nBlPar +7]

      return avTurb

    def forcedTransition(self, dataCont, marker):

      if marker < dataCont.bNodes[1]:
        avTurb = (dataCont.xGlobal[marker]-dataCont.xtrF[0])/(dataCont.xGlobal[marker]-dataCont.xGlobal[marker-1]) # Percentage of the subsection corresponding to a turbulent flow
      elif marker < dataCont.bNodes[2]:
        avTurb = (dataCont.xGlobal[marker]-dataCont.xtrF[1])/(dataCont.xGlobal[marker]-dataCont.xGlobal[marker-1]) # Percentage of the subsection corresponding to a turbulent flow
      else:
        print('Trying to impose transition in the wake')
      avLam = 1- avTurb # Percentage of the subsection corresponding to a laminar flow

      # Boundary condition
      dataCont.blPar[(marker-1)*dataCont.nBlPar + 7] = Closures.turbulentClosure(dataCont.u[(marker-1)*dataCont.nVar+0], dataCont.u[(marker-1)*dataCont.nVar+1],0.03, dataCont.blPar[(marker-1)*dataCont.nBlPar+0], dataCont.extPar[(marker-1)*dataCont.nExtPar+0],'airfoil')[6]
      dataCont.u[(marker-1)*dataCont.nVar + 4] = 0.7*avLam*dataCont.blPar[(marker-1)*dataCont.nBlPar +7]

      return avTurb

    def transitionBC(self, dataCont, marker):
      if marker < dataCont.bNodes[1]:
          xtr  = dataCont.xtr[0]
      elif dataCont.bNodes[1] <= marker < dataCont.bNodes[2]:
          xtr = dataCont.xtr[1]

      avTurb = (dataCont.xGlobal[marker]-dataCont.xtr[0])/(dataCont.xGlobal[marker]-dataCont.xGlobal[marker-1]) # Percentage of the subsection corresponding to a turbulent flow
      avLam = 1- avTurb # Percentage of the subsection corresponding to a laminar flow
      
      # Boundary condition
      dataCont.blPar[(marker-1)*dataCont.nBlPar + 7] = Closures.turbulentClosure(dataCont.u[(marker-1)*dataCont.nVar+0], dataCont.u[(marker-1)*dataCont.nVar+1],0.03, dataCont.blPar[(marker-1)*dataCont.nBlPar+0], dataCont.extPar[(marker-1)*dataCont.nExtPar+0],'airfoil')[6]
      dataCont.u[(marker-1)*dataCont.nVar + 4] = 0.7*avLam*dataCont.blPar[(marker-1)*dataCont.nBlPar +7]

      return avTurb